@@ -13,13 +13,12 @@ from unittest.mock import MagicMock, Mock
 
 import nbformat
 import pytest
-import xmltodict  # type: ignore
-from ipython_genutils.py3compat import string_types
+import xmltodict
 from jupyter_client import KernelManager
 from jupyter_client.kernelspec import KernelSpecManager
-from nbconvert.filters import strip_ansi  # type: ignore
+from nbconvert.filters import strip_ansi
 from nbformat import NotebookNode
-from testpath import modified_env  # type: ignore
+from testpath import modified_env
 from traitlets import TraitError
 
 from .. import NotebookClient, execute
@@ -41,6 +40,7 @@ hook_methods = [
     "on_cell_start",
     "on_cell_execute",
     "on_cell_complete",
+    "on_cell_executed",
     "on_cell_error",
     "on_notebook_start",
     "on_notebook_complete",
@@ -48,6 +48,32 @@ hook_methods = [
 ]
 
 
+def get_executor_with_hooks(nb=None, executor=None, async_hooks=False):
+    if async_hooks:
+        hooks = {key: AsyncMock() for key in hook_methods}
+    else:
+        hooks = {key: MagicMock() for key in hook_methods}
+    if nb is not None:
+        if executor is not None:
+            raise RuntimeError("Cannot pass nb and executor at the same time")
+        executor = NotebookClient(nb)
+    for k, v in hooks.items():
+        setattr(executor, k, v)
+    return executor, hooks
+
+
+EXECUTE_REPLY_OK = {
+    'parent_header': {'msg_id': 'fake_id'},
+    'content': {'status': 'ok', 'execution_count': 1},
+}
+EXECUTE_REPLY_ERROR = {
+    'parent_header': {'msg_id': 'fake_id'},
+    'content': {'status': 'error'},
+    'msg_type': 'execute_reply',
+    'header': {'msg_type': 'execute_reply'},
+}
+
+
 class AsyncMock(Mock):
     pass
 
@@ -127,7 +153,7 @@ async def async_run_notebook(filename, opts, resources=None):
     return input_nb, output_nb
 
 
-def prepare_cell_mocks(*messages, reply_msg=None):
+def prepare_cell_mocks(*messages_input, reply_msg=None):
     """
     This function prepares a executor object which has a fake kernel client
     to mock the messages sent over zeromq. The mock kernel client will return
@@ -135,7 +161,7 @@ def prepare_cell_mocks(*messages, reply_msg=None):
     callbacks. It also appends a kernel idle message to the end of messages.
     """
     parent_id = 'fake_id'
-    messages = list(messages)
+    messages = list(messages_input)
     # Always terminate messages with an idle to exit the loop
     messages.append({'msg_type': 'status', 'content': {'execution_state': 'idle'}})
 
@@ -215,7 +241,7 @@ def normalize_output(output):
     if 'image/svg+xml' in output.get('data', {}):
         output['data']['image/svg+xml'] = xmltodict.parse(output['data']['image/svg+xml'])
     for key, value in output.get('data', {}).items():
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             output['data'][key] = normalize_base64(value)
     if 'traceback' in output:
         tb = []
@@ -314,8 +340,10 @@ def test_parallel_notebooks(capfd, tmpdir):
             threading.Thread(target=run_notebook, args=(input_file.format(label=label), opts, res))
             for label in ("A", "B")
         ]
-        [t.start() for t in threads]
-        [t.join(timeout=2) for t in threads]
+        for t in threads:
+            t.start()
+        for t in threads:
+            t.join(timeout=2)
 
     captured = capfd.readouterr()
     assert filter_messages_on_error_output(captured.err) == ""
@@ -696,7 +724,7 @@ while True: continue
             self.assertNotEqual(call_count, 0, f'{method} was called')
 
     def test_process_message_wrapper(self):
-        outputs = []
+        outputs: list = []
 
         class WrappedPreProc(NotebookClient):
             def process_message(self, msg, cell, cell_index):
@@ -759,77 +787,79 @@ while True: continue
         filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
         with open(filename) as f:
             input_nb = nbformat.read(f, 4)
-        hooks = [MagicMock() for i in range(7)]
-        executor = NotebookClient(input_nb)
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(nb=input_nb)
         executor.execute()
-        for hook in hooks[:3]:
-            hook.assert_called_once()
-        hooks[3].assert_not_called()
-        for hook in hooks[4:6]:
-            hook.assert_called_once()
-        hooks[6].assert_not_called()
+        hooks["on_cell_start"].assert_called_once()
+        hooks["on_cell_execute"].assert_called_once()
+        hooks["on_cell_complete"].assert_called_once()
+        hooks["on_cell_executed"].assert_called_once()
+        hooks["on_cell_error"].assert_not_called()
+        hooks["on_notebook_start"].assert_called_once()
+        hooks["on_notebook_complete"].assert_called_once()
+        hooks["on_notebook_error"].assert_not_called()
 
     def test_error_execution_hook_error(self):
         filename = os.path.join(current_dir, 'files', 'Error.ipynb')
         with open(filename) as f:
             input_nb = nbformat.read(f, 4)
-        hooks = [MagicMock() for i in range(7)]
-        executor = NotebookClient(input_nb)
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(nb=input_nb)
         with pytest.raises(CellExecutionError):
             executor.execute()
-        for hook in hooks[:5]:
-            hook.assert_called_once()
-        hooks[6].assert_not_called()
+        hooks["on_cell_start"].assert_called_once()
+        hooks["on_cell_execute"].assert_called_once()
+        hooks["on_cell_complete"].assert_called_once()
+        hooks["on_cell_executed"].assert_called_once()
+        hooks["on_cell_error"].assert_called_once()
+        hooks["on_notebook_start"].assert_called_once()
+        hooks["on_notebook_complete"].assert_called_once()
+        hooks["on_notebook_error"].assert_not_called()
 
     def test_error_notebook_hook(self):
         filename = os.path.join(current_dir, 'files', 'Autokill.ipynb')
         with open(filename) as f:
             input_nb = nbformat.read(f, 4)
-        hooks = [MagicMock() for i in range(7)]
-        executor = NotebookClient(input_nb)
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(nb=input_nb)
         with pytest.raises(RuntimeError):
             executor.execute()
-        for hook in hooks[:3]:
-            hook.assert_called_once()
-        hooks[3].assert_not_called()
-        for hook in hooks[4:]:
-            hook.assert_called_once()
+        hooks["on_cell_start"].assert_called_once()
+        hooks["on_cell_execute"].assert_called_once()
+        hooks["on_cell_complete"].assert_called_once()
+        hooks["on_cell_executed"].assert_not_called()
+        hooks["on_cell_error"].assert_not_called()
+        hooks["on_notebook_start"].assert_called_once()
+        hooks["on_notebook_complete"].assert_called_once()
+        hooks["on_notebook_error"].assert_called_once()
 
     def test_async_execution_hook(self):
         filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
         with open(filename) as f:
             input_nb = nbformat.read(f, 4)
-        hooks = [AsyncMock() for i in range(7)]
-        executor = NotebookClient(input_nb)
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(nb=input_nb)
         executor.execute()
-        for hook in hooks[:3]:
-            hook.assert_called_once()
-        hooks[3].assert_not_called()
-        for hook in hooks[4:6]:
-            hook.assert_called_once()
-        hooks[6].assert_not_called()
+        hooks["on_cell_start"].assert_called_once()
+        hooks["on_cell_execute"].assert_called_once()
+        hooks["on_cell_complete"].assert_called_once()
+        hooks["on_cell_executed"].assert_called_once()
+        hooks["on_cell_error"].assert_not_called()
+        hooks["on_notebook_start"].assert_called_once()
+        hooks["on_notebook_complete"].assert_called_once()
+        hooks["on_notebook_error"].assert_not_called()
 
     def test_error_async_execution_hook(self):
         filename = os.path.join(current_dir, 'files', 'Error.ipynb')
         with open(filename) as f:
             input_nb = nbformat.read(f, 4)
-        hooks = [AsyncMock() for i in range(7)]
-        executor = NotebookClient(input_nb)
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(nb=input_nb)
         with pytest.raises(CellExecutionError):
-            executor.execute().execute()
-        for hook in hooks[:5]:
-            hook.assert_called_once()
-        hooks[6].assert_not_called()
+            executor.execute()
+        hooks["on_cell_start"].assert_called_once()
+        hooks["on_cell_execute"].assert_called_once()
+        hooks["on_cell_complete"].assert_called_once()
+        hooks["on_cell_executed"].assert_called_once()
+        hooks["on_cell_error"].assert_called_once()
+        hooks["on_notebook_start"].assert_called_once()
+        hooks["on_notebook_complete"].assert_called_once()
+        hooks["on_notebook_error"].assert_not_called()
 
 
 class TestRunCell(NBClientTestsBase):
@@ -1617,14 +1647,18 @@ class TestRunCell(NBClientTestsBase):
 
     @prepare_cell_mocks()
     def test_cell_hooks(self, executor, cell_mock, message_mock):
-        hooks = [MagicMock() for i in range(7)]
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(executor=executor)
         executor.execute_cell(cell_mock, 0)
-        for hook in hooks[:3]:
-            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
-        for hook in hooks[4:]:
-            hook.assert_not_called()
+        hooks["on_cell_start"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_execute"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_complete"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_executed"].assert_called_once_with(
+            cell=cell_mock, cell_index=0, execute_reply=EXECUTE_REPLY_OK
+        )
+        hooks["on_cell_error"].assert_not_called()
+        hooks["on_notebook_start"].assert_not_called()
+        hooks["on_notebook_complete"].assert_not_called()
+        hooks["on_notebook_error"].assert_not_called()
 
     @prepare_cell_mocks(
         {
@@ -1640,15 +1674,21 @@ class TestRunCell(NBClientTestsBase):
         },
     )
     def test_error_cell_hooks(self, executor, cell_mock, message_mock):
-        hooks = [MagicMock() for i in range(7)]
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(executor=executor)
         with self.assertRaises(CellExecutionError):
             executor.execute_cell(cell_mock, 0)
-        for hook in hooks[:4]:
-            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
-        for hook in hooks[5:]:
-            hook.assert_not_called()
+        hooks["on_cell_start"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_execute"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_complete"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_executed"].assert_called_once_with(
+            cell=cell_mock, cell_index=0, execute_reply=EXECUTE_REPLY_ERROR
+        )
+        hooks["on_cell_error"].assert_called_once_with(
+            cell=cell_mock, cell_index=0, execute_reply=EXECUTE_REPLY_ERROR
+        )
+        hooks["on_notebook_start"].assert_not_called()
+        hooks["on_notebook_complete"].assert_not_called()
+        hooks["on_notebook_error"].assert_not_called()
 
     @prepare_cell_mocks(
         reply_msg={
@@ -1660,25 +1700,31 @@ class TestRunCell(NBClientTestsBase):
     )
     def test_non_code_cell_hooks(self, executor, cell_mock, message_mock):
         cell_mock = NotebookNode(source='"foo" = "bar"', metadata={}, cell_type='raw', outputs=[])
-        hooks = [MagicMock() for i in range(7)]
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(executor=executor)
         executor.execute_cell(cell_mock, 0)
-        for hook in hooks[:1]:
-            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
-        for hook in hooks[1:]:
-            hook.assert_not_called()
+        hooks["on_cell_start"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_execute"].assert_not_called()
+        hooks["on_cell_complete"].assert_not_called()
+        hooks["on_cell_executed"].assert_not_called()
+        hooks["on_cell_error"].assert_not_called()
+        hooks["on_notebook_start"].assert_not_called()
+        hooks["on_notebook_complete"].assert_not_called()
+        hooks["on_notebook_error"].assert_not_called()
 
     @prepare_cell_mocks()
     def test_async_cell_hooks(self, executor, cell_mock, message_mock):
-        hooks = [AsyncMock() for i in range(7)]
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(executor=executor, async_hooks=True)
         executor.execute_cell(cell_mock, 0)
-        for hook in hooks[:3]:
-            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
-        for hook in hooks[4:]:
-            hook.assert_not_called()
+        hooks["on_cell_start"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_execute"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_complete"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_executed"].assert_called_once_with(
+            cell=cell_mock, cell_index=0, execute_reply=EXECUTE_REPLY_OK
+        )
+        hooks["on_cell_error"].assert_not_called()
+        hooks["on_notebook_start"].assert_not_called()
+        hooks["on_notebook_complete"].assert_not_called()
+        hooks["on_notebook_error"].assert_not_called()
 
     @prepare_cell_mocks(
         {
@@ -1694,12 +1740,18 @@ class TestRunCell(NBClientTestsBase):
         },
     )
     def test_error_async_cell_hooks(self, executor, cell_mock, message_mock):
-        hooks = [AsyncMock() for i in range(7)]
-        for executor_hook, hook in zip(hook_methods, hooks):
-            setattr(executor, executor_hook, hook)
+        executor, hooks = get_executor_with_hooks(executor=executor, async_hooks=True)
         with self.assertRaises(CellExecutionError):
             executor.execute_cell(cell_mock, 0)
-        for hook in hooks[:4]:
-            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
-        for hook in hooks[4:]:
-            hook.assert_not_called()
+        hooks["on_cell_start"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_execute"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_complete"].assert_called_once_with(cell=cell_mock, cell_index=0)
+        hooks["on_cell_executed"].assert_called_once_with(
+            cell=cell_mock, cell_index=0, execute_reply=EXECUTE_REPLY_ERROR
+        )
+        hooks["on_cell_error"].assert_called_once_with(
+            cell=cell_mock, cell_index=0, execute_reply=EXECUTE_REPLY_ERROR
+        )
+        hooks["on_notebook_start"].assert_not_called()
+        hooks["on_notebook_complete"].assert_not_called()
+        hooks["on_notebook_error"].assert_not_called()
